from transfer import Transmitter
from sender import *
from receiver import *
from defines import *

from sys import argv,exit
import threading
import random
import math
from socket import *
from time import sleep
from XmlMessage import XmlMessage

# necessary for retrieving information about personal ip and internet interface
import fcntl
import struct
import array

class Node:
	
	def __init__(self,filename,my_id):
		### Declarations
		
		### Initializations
		self.transmitter = Transmitter(filename)
		self.ip_list = self.transmitter.ip_list
		self.get_info(my_id)
		
		self.state = FREE 
		self.state_lock = threading.Lock()
		
		# valid for both of them
		self.seq_lock = threading.Lock()
		

		self.seq =  1
		# initial test suceed
		self.hiseq = 1

		# list containing authorizations for each site 
		# 1 means we're authorized. Initially, we're not
		self.auth = map(lambda x: 0, range(len(self.transmitter.ip_list)))
		self.auth_lock = threading.Condition()
		self.auth[self.myno-1] = 1
		
		# This conditions only holds durint initialization
		self.deferred_auth = map(lambda x: 0, range(len(self.transmitter.ip_list)))
		self.deferred_auth[self.myno-1] = 0
		
		
		# NEVER, EVER, call this lock with auth_lock held	
		self.resource_lock = threading.Lock()
		global S
		global T
		
		S = self.transmitter.S
		T = self.transmitter.T
		
		self.sender = Sender(self)
		self.receiver = Receiver(self)
		
		self.receiver.start()
		
		sleep(5)
		
		# shoot timer for resource 
		self.request_resource()
		
		
		
	def get_info(self, id):
		# sets personal attributes such as node#, port and ip
		my_id = int(id)-1
		[my_ip, port] = self.ip_list[my_id]
		
		valid_ip = filter(lambda ip: ip == my_ip, map((lambda interface:\
			self.get_ip_address(interface)), self.all_interfaces()))
				
		if not valid_ip:
			debug("ERROR - Given IP is not valid for available internet interfaces.")
			exit(-1)
			
		self.myno = int(id)
		self.port = port
		self.my_ip = my_ip
		
	def get_ip_address(self, ifname):
		#  get the IP address associated with a network interface
		s = socket(AF_INET, SOCK_DGRAM)
		return inet_ntoa(fcntl.ioctl(s.fileno(),0x8915,struct.pack('256s', ifname[:15]))[20:24])

	def all_interfaces(self):
		# get names of all "up" network interfaces
		max_possible = 10  # arbitrary. raise if needed.
		bytes = max_possible * 32
		s = socket(AF_INET, SOCK_DGRAM)
		names = array.array('B', '\0' * bytes)
		# 0x8912 -> SIOCGIFCONF
		outbytes = struct.unpack('iL', fcntl.ioctl(s.fileno(), 0x8912, \
		struct.pack('iL', bytes, names.buffer_info()[0])))[0]
		namestr = names.tostring()
		return [namestr[i:i+32].split('\0', 1)[0] for i in range(0, outbytes, 32)]
	
	def request_resource(self):
		timeout = (random.Random().random()*10)%S
		timer = threading.Timer(timeout,self.__do_request_resource)
		timer.start()
		

	def __do_request_resource(self):

		self.state_lock.acquire()
		self.state = WAITING
		self.state_lock.release()

		self.seq_lock.acquire()
		self.seq = self.hiseq + 1
		self.seq_lock.release()
		

		self.auth_lock.acquire()
		for node,st in enumerate(self.auth):
			if (node + 1 != self.myno) and (not st):
				self.sender.send_request(node+1)	
		while 0 in self.auth:
			self.auth_lock.wait()
		self.auth_lock.release()
		self.use_resource()

	def use_resource (self):
		self.state_lock.acquire()
		self.state = USING
		self.state_lock.release()
		
		print "USING RESOURCE."
		self.transmitter.requestResource()
		
		# Dorme um tempo aleatorio, sem fazer nada
		self.resource_lock.acquire()
		
		timeout = (random.Random().random()*10)%T
		sleep(timeout)
		self.resource_lock.release()
		
		# after using for a while and releasing, wait a bit
		# and try to use resource again
		self.release_resource()
		
		# then... we re-request
		self.request_resource()

	def release_resource (self):
		print "RELEASED RESOURCE.\n"
		
		self.transmitter.releaseResource()
		
		self.auth_lock.acquire()
		for node,st in enumerate(self.deferred_auth):
			# We'll never put ourselves in the deferred list,
			# so, no need to test
			if st == 1:
				self.auth[node] = 0 # not authorized
				self.sender.send_reply(node+1)
				self.deferred_auth[node] = 0 # no need to answer this
		self.auth_lock.notifyAll()
		self.auth_lock.release()
		
		self.state_lock.acquire()
		self.state = FREE
		self.state_lock.release()

if __name__ == "__main__":

	if len(argv) < 2:
		print "Please inform a configuration file"
		exit(-1)
	if len(argv) < 3:
		print "Please inform node id"
		exit(-1)
	else:
		filename = argv[1]
		node_id = argv[2]
		n = Node(filename,node_id)
